from sqlalchemy import Column, \
    Integer, \
    String, \
    BigInteger, \
    ForeignKey
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship

# noinspection PyUnresolvedReferences
from .users import User


class Discord:
    __tablename__ = "discord"

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"), nullable=False)

    @declared_attr
    def user(self):
        return relationship("User", backref="discord")

    @declared_attr
    def discord_id(self):
        return Column(BigInteger, primary_key=True)

    @declared_attr
    def username(self):
        return Column(String)

    @declared_attr
    def discriminator(self):
        return Column(String)

    @declared_attr
    def avatar_url(self):
        return Column(String)

    def __repr__(self):
        return f"<Discord {str(self)}>"

    def __str__(self):
        return f"[c]discord:{self.full_username()}[/c]"

    def full_username(self):
        return f"{self.username}#{self.discriminator}"



from .errors import ConfigurationError


class ConfigDict(dict):
    def __missing__(self, key):
        raise ConfigurationError(f"Missing config key '{key}'")

    @classmethod
    def convert(cls, item):
        if isinstance(item, dict):
            cd = ConfigDict()
            for key in item:
                cd[key] = cls.convert(item[key])
            return cd
        elif isinstance(item, list):
            nl = []
            for obj in item:
                nl.append(cls.convert(obj))
        else:
            return item



from .apidata import ApiData
from .apierrors import \
    ApiError, \
    NotFoundError, \
    ForbiddenError, \
    BadRequestError, \
    ParameterError, \
    MissingParameterError, \
    InvalidParameterError, \
    MethodNotImplementedError, \
    UnsupportedError
from .apistar import ApiStar
from .jsonapi import api_response, api_success, api_error
from .magic import magic

__all__ = [
    "ApiStar",
    "api_response",
    "api_success",
    "api_error",
    "ApiData",
    "ApiError",
    "MissingParameterError",
    "NotFoundError",
    "ApiError",
    "NotFoundError",
    "ForbiddenError",
    "BadRequestError",
    "ParameterError",
    "MissingParameterError",
    "InvalidParameterError",
    "MethodNotImplementedError",
    "UnsupportedError",
    "magic",
]



import asyncio as aio
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..serf import Serf
    from ..constellation import Constellation


class HeraldEvent:
    """A remote procedure call triggered by a :mod:`royalnet.herald` request."""

    name = NotImplemented
    """The event_name that will trigger this event."""

    def __init__(self, parent: Union["Serf", "Constellation"], config):
        self.parent: Union["Serf", "Constellation"] = parent
        self.config = config

    @property
    def alchemy(self):
        """A shortcut for :attr:`.parent.alchemy`."""
        return self.parent.alchemy

    @property
    def loop(self) -> aio.AbstractEventLoop:
        """A shortcut for :attr:`.parent.loop`."""
        return self.parent.loop

    async def run(self, **kwargs):
        raise NotImplementedError()



import asyncio as aio
import contextlib
import logging
import uuid
from dataclasses import dataclass
from typing import *

import telegram
import urllib3
from telegram.utils.request import Request as TRequest

import royalnet.backpack.tables as rbt
import royalnet.commands as rc
import royalnet.utils as ru
from .escape import escape
from ..serf import Serf

try:
    from sqlalchemy.orm.session import Session
except ImportError:
    Session = None

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class TelegramKeyCallback:
    command: rc.Command
    key: rc.KeyboardKey


class TelegramSerf(Serf):
    """A Serf that connects to `Telegram <https://telegram.org/>`_ as a bot."""
    interface_name = "telegram"
    prefix = "/"

    def __init__(self,
                 loop: aio.AbstractEventLoop,
                 alchemy_cfg: rc.ConfigDict,
                 herald_cfg: rc.ConfigDict,
                 sentry_cfg: rc.ConfigDict,
                 packs_cfg: rc.ConfigDict,
                 serf_cfg: rc.ConfigDict,
                 **_):
        if telegram is None:
            raise ImportError("'telegram' extra is not installed")

        super().__init__(loop=loop,
                         alchemy_cfg=alchemy_cfg,
                         herald_cfg=herald_cfg,
                         sentry_cfg=sentry_cfg,
                         packs_cfg=packs_cfg,
                         serf_cfg=serf_cfg)

        self.client = telegram.Bot(serf_cfg["token"],
                                   request=TRequest(serf_cfg["pool_size"],
                                                    read_timeout=serf_cfg["read_timeout"]))
        """The :class:`telegram.Bot` instance that will be used from the Serf."""

        self.update_offset: int = -100
        """The current `update offset <https://core.telegram.org/bots/api#getupdates>`_."""

        self.key_callbacks: Dict[str, TelegramKeyCallback] = {}

        self.MessageData: Type[rc.CommandData] = self.message_data_factory()
        self.CallbackData: Type[rc.CommandData] = self.callback_data_factory()

    @staticmethod
    async def api_call(f: Callable, *args, **kwargs) -> Optional:
        """Call a :class:`telegram.Bot` method safely, without getting a mess of errors raised.

        The method may return None if it was decided that the call should be skipped."""
        while True:
            try:
                return await ru.asyncify(f, *args, **kwargs)
            except telegram.error.TimedOut as error:
                log.debug(f"Timed out during {f.__qualname__} (retrying immediatly): {error}")
                continue
            except telegram.error.NetworkError as error:
                log.debug(f"Network error during {f.__qualname__} (skipping): {error}")
                break
            except telegram.error.Unauthorized as error:
                log.info(f"Unauthorized to run {f.__qualname__} (skipping): {error}")
                break
            except telegram.error.RetryAfter as error:
                log.warning(f"Rate limited during {f.__qualname__} (retrying in 15s): {error}")
                await aio.sleep(15)
                continue
            except urllib3.exceptions.HTTPError as error:
                log.warning(f"urllib3 HTTPError during {f.__qualname__} (retrying in 15s): {error}")
                await aio.sleep(15)
                continue
            except Exception as error:
                log.error(f"{error.__class__.__qualname__} during {f} (skipping): {error}")
                ru.sentry_exc(error)
                break
        return None

    def message_data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters
        class TelegramMessageData(rc.CommandData):
            def __init__(data,
                         command: rc.Command,
                         message: telegram.Message):
                super().__init__(command=command)
                data.message: telegram.Message = message

            async def reply(data, text: str):
                await self.api_call(data.message.chat.send_message,
                                    escape(text),
                                    parse_mode="HTML",
                                    disable_web_page_preview=True)

            async def reply_image(data, image: "BinaryIO", caption: Optional[str] = None) -> None:
                await self.api_call(data.message.chat.send_photo,
                                    photo=image,
                                    caption=escape(caption) if caption is not None else None,
                                    parse_mode="HTML",
                                    disable_web_page_preview=True)

            async def find_author(data,
                                  *,
                                  session,
                                  required: bool = False) -> Optional[rbt.User]:
                user: "telegram.User" = data.message.from_user
                TelegramT = data.alchemy.get(rbt.Telegram)
                result = await ru.asyncify(
                    session.query(TelegramT).filter(TelegramT.tg_id == user.id).one_or_none
                )
                if result is None and required:
                    raise rc.CommandError("You must be registered to use this command.")
                return result.user

            async def delete_invoking(data, error_if_unavailable=False) -> None:
                await self.api_call(data.message.delete)

            @contextlib.asynccontextmanager
            async def keyboard(data, text: str, keys: List[rc.KeyboardKey]):
                tg_rows = []
                key_uids = []
                for key in keys:
                    uid: str = str(uuid.uuid4())
                    key_uids.append(uid)
                    data.register_keyboard_key(uid, key)
                    tg_button: telegram.InlineKeyboardButton = telegram.InlineKeyboardButton(key.text,
                                                                                             callback_data=uid)
                    tg_row: List[telegram.InlineKeyboardButton] = [tg_button]
                    tg_rows.append(tg_row)
                tg_markup: telegram.InlineKeyboardMarkup = telegram.InlineKeyboardMarkup(tg_rows)
                message: telegram.Message = await self.api_call(data.message.chat.send_message,
                                                                escape(text),
                                                                parse_mode="HTML",
                                                                disable_web_page_preview=True,
                                                                reply_markup=tg_markup)
                yield message
                await self.api_call(message.edit_reply_markup, reply_markup=None)
                for uid in key_uids:
                    data.unregister_keyboard_key(uid)

            def register_keyboard_key(data, identifier: str, key: rc.KeyboardKey):
                self.key_callbacks[identifier] = TelegramKeyCallback(key=key, command=data.command)

            def unregister_keyboard_key(data, identifier: str):
                del self.key_callbacks[identifier]

        return TelegramMessageData

    def register_keyboard_key(self, identifier: str, key: rc.KeyboardKey, command: rc.Command):
        self.key_callbacks[identifier] = TelegramKeyCallback(key=key, command=command)

    def unregister_keyboard_key(self, identifier: str):
        del self.key_callbacks[identifier]

    def callback_data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters
        class TelegramKeyboardData(rc.CommandData):
            def __init__(data,
                         command: rc.Command,
                         cbq: telegram.CallbackQuery):
                super().__init__(command=command)
                data.cbq: telegram.CallbackQuery = cbq

            async def reply(data, text: str):
                await self.api_call(data.cbq.answer,
                                    escape(text))

            async def find_author(data,
                                  *,
                                  session,
                                  required: bool = False) -> Optional[rbt.User]:
                user: "telegram.User" = data.cbq.from_user
                TelegramT = data.alchemy.get(rbt.Telegram)
                result = await ru.asyncify(
                    session.query(TelegramT).filter(TelegramT.tg_id == user.id).one_or_none
                )
                if result is None and required:
                    raise rc.CommandError("You must be registered to use this command.")
                return result.user

        return TelegramKeyboardData

    async def answer_cbq(self, cbq, text, alert=False):
        await self.api_call(cbq.answer, text=text, show_alert=alert)

    async def handle_update(self, update: telegram.Update):
        """Delegate :class:`telegram.Update` handling to the correct message type submethod."""
        if update.message is not None:
            log.debug(f"Handling update as a message")
            await self.handle_message(update.message)
        elif update.edited_message is not None:
            log.debug(f"Update is a edited message, not doing anything")
        elif update.channel_post is not None:
            log.debug(f"Update is a channel post, not doing anything")
        elif update.edited_channel_post is not None:
            log.debug(f"Update is a channel edit, not doing anything")
        elif update.inline_query is not None:
            log.debug(f"Update is a inline query, not doing anything")
        elif update.chosen_inline_result is not None:
            log.debug(f"Update is a chosen inline result, not doing anything")
        elif update.callback_query is not None:
            log.debug(f"Handling update as a callback query")
            await self.handle_callback_query(update.callback_query)
        elif update.shipping_query is not None:
            log.debug(f"Update is a shipping query, not doing anything")
        elif update.pre_checkout_query is not None:
            log.debug(f"Update is a precheckout query, not doing anything")
        elif update.poll is not None:
            log.debug(f"Update is a poll, not doing anything")
        else:
            log.warning(f"Unknown update type: {update}")

    async def handle_message(self, message: telegram.Message):
        """What should be done when a :class:`telegram.Message` is received?"""
        text: str = message.text
        # Try getting the caption instead
        if text is None:
            text: str = message.caption
        # No text or caption, ignore the message
        if text is None:
            log.debug("Skipping message as it had no text or caption")
            return
        # Skip non-command updates
        if not text.startswith(self.prefix):
            log.debug(f"Skipping message as it didn't start with {self.prefix}")
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        command_name = command_text.replace(f"@{self.client.username}", "").lstrip(self.prefix).lower()
        log.debug(f"Parsed '{command_name}' as command name")
        # Find the command
        try:
            command = self.commands[command_name]
        except KeyError:
            # Skip the message
            log.debug(f"Skipping message as I could not find the command {command_name}")
            return
        # Send a typing notification
        log.debug(f"Sending typing notification")
        await self.api_call(message.chat.send_action, telegram.ChatAction.TYPING)
        # Prepare data
        # noinspection PyArgumentList
        data = self.MessageData(command=command, message=message)
        # Call the command
        log.debug(f"Calling {command}")
        await self.call(command, data, parameters)

    async def handle_callback_query(self, cbq: telegram.CallbackQuery):
        uid = cbq.data
        if uid not in self.key_callbacks:
            await self.api_call(cbq.answer, text="⚠️ This keyboard has expired.", show_alert=True)
            return
        cbd = self.key_callbacks[uid]
        # noinspection PyArgumentList
        data: rc.CommandData = self.CallbackData(command=cbd.command, cbq=cbq)
        await self.press(cbd.key, data)

    async def run(self):
        await super().run()
        while True:
            # Collect ended tasks
            self.tasks.collect()
            # Get the latest 100 updates
            log.debug("Getting updates...")
            last_updates: Optional[List[telegram.Update]] = await self.api_call(
                self.client.get_updates,
                offset=self.update_offset,
                timeout=60,
                read_latency=5.0
            )
            # Ensure a list was returned from the API call
            if not isinstance(last_updates, list):
                log.warning("Received invalid data from get_updates, sleeping for 60 seconds, hoping it fixes itself.")
                await aio.sleep(60)
                continue
            # Handle updates
            log.debug("Handling updates...")
            for update in last_updates:
                self.tasks.add(self.handle_update(update))
            # Recalculate offset
            try:
                self.update_offset = last_updates[-1].update_id + 1
            except IndexError:
                pass



import royalnet.constellation.api as rca
import royalnet.utils as ru
from ..tables import *


class ApiUserCreateStar(rca.ApiStar):
    path = "/api/user/create/v1"

    parameters = {
        "post": {
            "username": "The name of the user you are creating.",
            "password": "The password of the user you are creating.",
            "email": "(Optional) The email of the user you are creating.",
            "avatar_url": "(Optional) An URL pointing to the avatar of the user you are creating.",
        }
    }

    tags = ["user"]

    @rca.magic
    async def post(self, data: rca.ApiData) -> ru.JSON:
        """Create a new Royalnet account."""
        UserT = self.alchemy.get(User)

        username = data.str("username")
        password = data.str("password")
        email = data.str("email", optional=True)
        avatar_url = data.str("avatar_url", optional=True)

        # Check if the username is used by an user or an alias
        user = await User.find(self.alchemy, data.session, username)
        if user is not None:
            raise rca.InvalidParameterError("An user with that username or alias already exists.")

        # Create the user
        user = UserT(
            username=username,
            email=email,
            avatar_url=avatar_url
        )
        data.session.add(user)
        user.set_password(password)
        user.add_alias(self.alchemy, username)
        await data.session_commit()

        return user.json()



import abc
from typing import *

from starlette.requests import Request
from starlette.responses import Response

if TYPE_CHECKING:
    from .constellation import Constellation
    from ..commands import ConfigDict
    from ..alchemy import Alchemy
    import sqlalchemy.orm.session


class Star(metaclass=abc.ABCMeta):
    """A Star is a class representing a part of the website.

    It shouldn't be used directly: please use :class:`PageStar` and :class:`ExceptionStar` instead!"""

    def __init__(self, constellation: "Constellation", config: "ConfigDict"):
        self.constellation: "Constellation" = constellation
        self.config: "ConfigDict" = config

    @abc.abstractmethod
    async def page(self, request: Request) -> Response:
        """The function generating the :class:`~starlette.Response` to a web :class:`~starlette.Request`.

        If it raises an error, the corresponding :class:`ExceptionStar` will be used to handle the request instead."""
        raise NotImplementedError()

    @property
    def alchemy(self) -> "Alchemy":
        """A shortcut for the :class:`~royalnet.alchemy.Alchemy` of the :class:`Constellation`."""
        return self.constellation.alchemy

    # noinspection PyPep8Naming
    @property
    def Session(self) -> "sqlalchemy.orm.session.Session":
        """A shortcut for the :class:`~royalnet.alchemy.Alchemy` :class:`Session` of the :class:`Constellation`."""
        return self.constellation.alchemy.Session

    @property
    def session_acm(self):
        """A shortcut for :func:`.alchemy.session_acm` of the :class:`Constellation`."""
        return self.constellation.alchemy.session_acm

    def __repr__(self):
        return f"<{self.__class__.__qualname__}>"



# Imports go here!
from .api_auth_login_royalnet import ApiAuthLoginRoyalnetStar
from .api_auth_token import ApiAuthTokenStar
from .api_royalnet_version import ApiRoyalnetVersionStar
from .api_user_create import ApiUserCreateStar
from .api_user_find import ApiUserFindStar
from .api_user_get import ApiUserGetStar
from .api_user_list import ApiUserListStar
from .api_user_passwd import ApiUserPasswd
from .docs import DocsStar

# Enter the PageStars of your Pack here!
available_page_stars = [
    ApiRoyalnetVersionStar,
    ApiAuthLoginRoyalnetStar,
    ApiUserPasswd,
    ApiAuthTokenStar,
    ApiUserGetStar,
    ApiUserListStar,
    ApiUserFindStar,
    ApiUserCreateStar,
    DocsStar,
]

# Don't change this, it should automatically generate __all__
__all__ = [star.__name__ for star in available_page_stars]



"""The subpackage providing all functions and classes to handle communication between process (even over the Internet).

It is based on :mod:`websockets`.

It requires the ``herald`` extra to be installed.

You can install it with: ::

    pip install royalnet[herald]

"""

from .broadcast import Broadcast
from .config import Config
from .errors import *
from .link import Link
from .package import Package
from .request import Request
from .response import Response, ResponseSuccess, ResponseFailure
from .server import Server

__all__ = [
    "Config",
    "HeraldError",
    "ConnectionClosedError",
    "LinkError",
    "InvalidServerResponseError",
    "ServerError",
    "Link",
    "Package",
    "Request",
    "Response",
    "ResponseSuccess",
    "ResponseFailure",
    "Server",
    "Broadcast",
]



import datetime
from typing import *

from sqlalchemy import and_

import royalnet.constellation.api as rca
import royalnet.utils as ru
from ..tables.tokens import Token


class ApiUserPasswd(rca.ApiStar):
    path = "/api/user/passwd/v1"

    tags = ["user"]

    parameters = {
        "put": {
            "new_password": "The password you want to set."
        }
    }

    auth = {
        "put": True
    }

    requires_auth = True

    @rca.magic
    async def put(self, data: rca.ApiData) -> ru.JSON:
        """Change the password of the currently logged in user.

        This method also revokes all the issued tokens for the user."""
        TokenT = self.alchemy.get(Token)
        token = await data.token()
        user = token.user
        user.set_password(data["new_password"])
        tokens: List[Token] = await ru.asyncify(
            data.session
                .query(self.alchemy.get(Token))
                .filter(and_(TokenT.user == user, TokenT.expiration >= datetime.datetime.now()))
                .all
        )
        for t in tokens:
            if t.token != token.token:
                t.expired = True
        await data.session_commit()
        return {
            "revoked_tokens": len(tokens) - 1
        }



import logging
from typing import *

import royalnet.utils as ru
from royalnet.backpack.tables.tokens import Token
from royalnet.backpack.tables.users import User
from .apierrors import *

log = logging.getLogger(__name__)


class ApiData(dict):
    def __init__(self, data, star):
        super().__init__(data)
        self.star = star
        self._session = None

    def __missing__(self, key):
        raise MissingParameterError(f"Missing '{key}'")

    def str(self, key, optional=False) -> Optional[str]:
        if optional:
            return self.get(key)
        else:
            return self[key]

    def int(self, key, optional=False) -> Optional[int]:
        value = self.str(key, optional)
        if value is None:
            return None
        try:
            return int(value)
        except ValueError:
            raise BadRequestError(f"Could not parse the value `{value}` as an int.")

    def float(self, key, optional=False) -> Optional[float]:
        value = self.str(key, optional)
        if value is None:
            return None
        try:
            return float(value)
        except ValueError:
            raise BadRequestError(f"Could not parse the value `{value}` as a float.")

    _bool_values = {
        "true": True,
        "t": True,
        "yes": True,
        "y": True,
        "false": False,
        "f": False,
        "no": False,
        "n": False
    }

    def bool(self, key, optional=False) -> Optional[bool]:
        value = self.str(key, optional)
        if value is None:
            return None
        value = value.lower()
        try:
            return self._bool_values[value]
        except KeyError:
            raise BadRequestError(f"Could not parse the value `{value}` as a bool.")

    async def token(self) -> Token:
        token = await Token.find(self.star.alchemy, self.session, self["token"])
        if token is None:
            raise ForbiddenError("'token' is invalid")
        if token.expired:
            raise ForbiddenError("Login token has expired")
        return token

    async def user(self) -> User:
        return (await self.token()).user

    @property
    def session(self):
        if self._session is None:
            if self.star.alchemy is None:
                raise UnsupportedError("'alchemy' is not enabled on this Royalnet instance")
            log.debug("Creating Session...")
            self._session = self.star.alchemy.Session()
        return self._session

    async def session_commit(self):
        """Asyncronously commit the :attr:`.session` of this object."""
        if self._session:
            log.warning("Session had to be created to be committed")
        # noinspection PyUnresolvedReferences
        log.debug("Committing Session...")
        await ru.asyncify(self.session.commit)

    async def session_close(self):
        """Asyncronously close the :attr:`.session` of this object."""
        if self._session is not None:
            log.debug("Closing Session...")
            await ru.asyncify(self._session.close)



import abc
import asyncio as aio
from typing import *

from .commandargs import CommandArgs
from .commanddata import CommandData

if TYPE_CHECKING:
    from .configdict import ConfigDict
    from ..serf import Serf
    from ..alchemy import Alchemy


class Command(metaclass=abc.ABCMeta):
    name: str = NotImplemented
    """The main name of the command.
    
    Example:
        To be able to call ``/example`` on Telegram, the name should be ``"example"``."""

    aliases: List[str] = []
    """A list of possible aliases for a command.
    
    Example:
        To be able to call ``/e`` as an alias for ``/example``, one should set aliases to ``["e"]``."""

    description: str = NotImplemented
    """A small description of the command, to be displayed when the command is being autocompleted."""

    syntax: str = ""
    """The syntax of the command, to be displayed when a :py:exc:`InvalidInputError` is raised,
     in the format ``(required_arg) [optional_arg]``."""

    def __init__(self, serf: "Serf", config: "ConfigDict"):
        self.serf: "Serf" = serf
        self.config: "ConfigDict" = config

    def __str__(self):
        return f"[c]{self.serf.prefix}{self.name}[/c]"

    @property
    def alchemy(self) -> "Alchemy":
        """A shortcut for :attr:`.serf.alchemy`."""
        return self.serf.alchemy

    @property
    def session_acm(self):
        """A shortcut for :attr:`.alchemy.session_acm`."""
        return self.alchemy.session_acm

    @property
    def loop(self) -> aio.AbstractEventLoop:
        """A shortcut for :attr:`.serf.loop`."""
        return self.serf.loop

    @abc.abstractmethod
    async def run(self, args: CommandArgs, data: CommandData) -> None:
        raise NotImplementedError()



from starlette.responses import *

import royalnet.backpack.tables as rbt
import royalnet.constellation.api as rca
import royalnet.utils as ru


class ApiUserListStar(rca.ApiStar):
    path = "/api/user/list/v1"

    tags = ["user"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get a list of all Royalnet users."""
        users: typing.List[rbt.User] = await ru.asyncify(data.session.query(self.alchemy.get(rbt.User)).all)
        return [user.json() for user in users]



import re


def strip_tabs(s: str) -> str:
    # https://github.com/Steffo99/bluelib/blob/pls-work/src/utils/stripTabs.js

    indent_regex = re.compile(r"^[ \t]+")

    rows = list(filter(lambda r: r != "", s.split("\n")))

    match = None
    for row in rows:
        match = re.match(indent_regex, row)
        if match is not None:
            break

    if match is None:
        start = 0
    else:
        start = len(match.group(0))

    return "\n".join(map(lambda r: r[start:], rows))



from .asyncify import asyncify
from .formatters import andformat, underscorize, ytdldateformat, numberemojiformat, ordinalformat
from .log import init_logging
from .multilock import MultiLock
from .royalnetprocess import RoyalnetProcess
from .royaltyping import JSON
from .sentry import init_sentry, sentry_exc, sentry_wrap, sentry_async_wrap
from .sleep_until import sleep_until
from .strip_tabs import strip_tabs
from .taskslist import TaskList
from .urluuid import to_urluuid, from_urluuid

__all__ = [
    "asyncify",
    "sleep_until",
    "andformat",
    "underscorize",
    "ytdldateformat",
    "numberemojiformat",
    "ordinalformat",
    "to_urluuid",
    "from_urluuid",
    "MultiLock",
    "init_sentry",
    "sentry_exc",
    "sentry_wrap",
    "sentry_async_wrap",
    "init_logging",
    "JSON",
    "strip_tabs",
    "TaskList",
    "RoyalnetProcess",
]



import contextlib
import io
import logging
from typing import *

from royalnet.backpack.tables.users import User
from .errors import *

if TYPE_CHECKING:
    from .keyboardkey import KeyboardKey
    from .command import Command

log = logging.getLogger(__name__)


class CommandData:
    def __init__(self, command: "Command"):
        self.command: "Command" = command

    @property
    def loop(self):
        return self.command.loop

    @property
    def alchemy(self):
        """A shortcut for :attr:`.command.alchemy`."""
        return self.command.alchemy

    @property
    def session_acm(self):
        """A shortcut for :attr:`.alchemy.session_acm`."""
        return self.alchemy.session_acm

    async def reply(self, text: str) -> None:
        """Send a text message to the channel where the call was made.

        Parameters:
             text: The text to be sent, possibly formatted in the weird undescribed markup that I'm using."""
        raise UnsupportedError(f"'{self.reply.__name__}' is not supported")

    async def reply_image(self, image: io.IOBase, caption: Optional[str] = None) -> None:
        """Send an image (with optionally a caption) to the channel where the call was made.

        Parameters:
            image: The bytes of the image to send.
            caption: The caption to attach to the image."""
        raise UnsupportedError(f"'{self.reply_image.__name__}' is not supported")

    async def delete_invoking(self, error_if_unavailable: bool = False) -> None:
        """Delete the invoking message, if supported by the interface.

        The invoking message is the message send by the user that contains the command.

        Parameters:
            error_if_unavailable: if True, raise an exception if the message cannot been deleted."""
        if error_if_unavailable:
            raise UnsupportedError(f"'{self.delete_invoking.__name__}' is not supported")

    async def find_author(self, *, session, required: bool = False) -> Optional["User"]:
        """Try to find the identifier of the user that sent the message.
        That probably means, the database row identifying the user.

        Parameters:
            session: the session that the user should be returned from.
            required: Raise an exception if this is True and the call has no author.
        """
        raise UnsupportedError(f"'{self.find_author.__name__}' is not supported")

    async def find_user(self, identifier: Union[str, int], *, session, required: bool = False) -> Optional["User"]:
        """Find the User having a specific identifier.

        Parameters:
            identifier: the identifier to search for.
            session: the session that the user should be returned from.
            required: Raise an exception if this is True and no user was found..
        """
        user: Optional["User"] = await User.find(alchemy=self.alchemy, session=session, identifier=identifier)
        if required and user is None:
            raise InvalidInputError(f"User '{identifier}' was not found.")
        return user

    @contextlib.asynccontextmanager
    async def keyboard(self, text, keys: List["KeyboardKey"]):
        yield
        raise UnsupportedError(f"{self.keyboard.__name__} is not supported")

    @classmethod
    def register_keyboard_key(cls, identifier: str, key: "KeyboardKey"):
        raise UnsupportedError(f"{cls.register_keyboard_key.__name__} is not supported")

    @classmethod
    def unregister_keyboard_key(cls, identifier: str):
        raise UnsupportedError(f"{cls.unregister_keyboard_key.__name__} is not supported")



"""A :class:`Serf` implementation for Discord.

It requires (obviously) the ``discord`` extra to be installed.

Install it with: ::

    pip install royalnet[discord]

"""

from .discordserf import DiscordSerf
from .escape import escape

__all__ = [
    "escape",
    "DiscordSerf",
]
